struct PostImputingMatrix{T <: Number, C <: AbstractVector{T}, U <: AbstractMatrix{T}} <: AbstractMatrix{T}
    W::U
    ψ::C
end

Flux.@functor PostImputingMatrix

PostImputingMatrix(W::AbstractMatrix{T}) where T = PostImputingMatrix(W, zeros(T, size(W, 1)))

Flux.@forward PostImputingMatrix.W Base.size, Base.length, Base.getindex, Base.setindex!, Base.firstindex, Base.lastindex

Base.vcat(As::PostImputingMatrix...) = PostImputingMatrix(vcat((A.W for A in As)...), vcat((A.ψ for A in As)...))
function Base.hcat(As::PostImputingMatrix...)
    ArgumentError("It doesn't make sense to hcat PostImputingMatrices") |> throw
end

A::PostImputingMatrix * b::AbstractVector = (_check_mul(A, b); _mul(A, b))
Zygote.@adjoint A::PostImputingMatrix * b::AbstractVector = (_check_mul(A, b); Zygote.pullback(_mul, A, b))
A::PostImputingMatrix * B::AbstractMatrix = (_check_mul(A, B); _mul(A, B))
Zygote.@adjoint A::PostImputingMatrix * B::AbstractMatrix = (_check_mul(A, B); Zygote.pullback(_mul, A, B))
A::PostImputingMatrix * b::MaybeHotVector = (_check_mul(A, b); _mul(A, b))
Zygote.@adjoint A::PostImputingMatrix * b::MaybeHotVector = (_check_mul(A, b); Zygote.pullback(_mul, A, b))
A::PostImputingMatrix * B::MaybeHotMatrix = (_check_mul(A, B); _mul(A, B))
Zygote.@adjoint A::PostImputingMatrix * B::MaybeHotMatrix = (_check_mul(A, B); Zygote.pullback(_mul, A, B))
A::PostImputingMatrix * B::NGramMatrix = (_check_mul(A, B); _mul(A, B))
Zygote.@adjoint A::PostImputingMatrix * B::NGramMatrix = (_check_mul(A, B); Zygote.pullback(_mul, A, B))

_mul(A::PostImputingMatrix, B::AbstractVecOrMat) = A.W * B
_mul(A::PostImputingMatrix, b::MaybeHotVector{Missing}) = A.ψ
_mul(A::PostImputingMatrix, b::MaybeHotVector{<:Integer}) = A.W * b
_mul(A::PostImputingMatrix, B::MaybeHotMatrix{Missing}) = repeat(A.ψ, 1, size(B, 2))
_mul(A::PostImputingMatrix, B::MaybeHotMatrix{<:Integer}) = A.W * B
_mul(A::PostImputingMatrix, B::MaybeHotMatrix{Maybe{T}}) where {T <: Integer} =
    _mul_maybe_hot(A.W, A.ψ, B.I)
_mul(A::PostImputingMatrix, B::NGramMatrix{<:Sequence}) = A.W * B
_mul(A::PostImputingMatrix, B::NGramMatrix{Missing}) = repeat(A.ψ, 1, size(B, 2))
_mul(A::PostImputingMatrix, B::NGramMatrix{Maybe{T}}) where {T <: Sequence} =
    _mul_ngram(A.W, A.ψ, B.s, B.n, B.b, B.m)

_mul_maybe_hot(W, ψ, I) = _impute_maybe_hot(W, ψ, I)[1]
function rrule(::typeof(_mul_maybe_hot), W, ψ, I)
    C, m = _impute_maybe_hot(W, ψ, I)
    function dW_thunk(Δ)
        dW = zero(W)
        @inbounds for (k, i) in enumerate(I)
            if !ismissing(i)
                @views dW[:, i] .+= Δ[:, k]
            end
        end
        dW
    end
    C, Δ -> (NO_FIELDS, @thunk(dW_thunk(Δ)), @thunk(sum(view(Δ, :, m), dims=2)), DoesNotExist())
end
function _impute_maybe_hot(W, ψ, I)
    m = trues(length(I))
    C = similar(ψ, size(W, 1), length(I))
    C .= ψ
    @inbounds for (k, i) in enumerate(I)
        if !ismissing(I[k])
            m[k] = false
            C[:, k] .= @view W[:, i]
        end
    end
    C, m
end

# TODO rewrite this to less parameters once Zygote allows for composite grads
_mul_ngram(W, ψ, S, n, b, m) = _impute_ngram(W, ψ, S, n, b, m)
function rrule(::typeof(_mul_ngram), W, ψ, S, n, b, m)
    function dW_thunk(Δ)
        dW = zero(W)
        z = _init_z(n, b)
        for (k, s) in enumerate(S)
            _dA_mul_vec!(Δ, k, dW, z, s, n, b, m)
        end
        dW
    end
    _impute_ngram(W, ψ, S, n, b, m), Δ -> (NO_FIELDS, @thunk(dW_thunk(Δ)),
                                  @thunk(sum(view(Δ, :, ismissing.(S)), dims=2)),
                                  fill(DoesNotExist(), 4)...)
end

function _impute_ngram(W, ψ, S, n, b, m)
    C = zeros(eltype(ψ), size(W, 1), length(S))
    z = _init_z(n, b)
    for (k, s) in enumerate(S)
        _mul_vec!(C, k, W, z, s, n, b, m, ψ)
    end
    C
end
